/*
 * Copyright (C) 2011 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by Michal Hruby <michal.hruby@canonical.com>
 *
 */

public class Main
{
  static bool remote_scope_test = false;
  public static int main (string[] args)
  {
    if ("--with-remote-scope" in args) remote_scope_test = true;

    if (remote_scope_test)
    {
      Environment.set_variable ("LIBUNITY_LENS_DIRECTORY",
                                Config.TESTDIR + "/data", true);
    }
    else
    {
      Environment.set_variable ("LIBUNITY_LENS_DIRECTORY",
                                Config.TESTDIR, true);
    }

    Test.init (ref args);

    Test.add_data_func ("/Unit/Lens/Export", test_lens_export);
    if (remote_scope_test)
    {
      Test.add_data_func ("/Integration/RemoteScope/Initialize", test_remote_scope_init);
    }
    Test.add_data_func ("/Unit/Lens/Search", test_lens_search);
    Test.add_data_func ("/Unit/LocalScope/Initialize", test_local_scope_init);
    Test.add_data_func ("/Unit/LocalScope/SearchChanged", test_local_scope_search);
    Test.add_data_func ("/Unit/Lens/ReturnAfterScopeFinish", test_lens_return_after_scope_finish);
    Test.add_data_func ("/Unit/Lens/TwoSearches", test_lens_two_searches);
    Test.add_data_func ("/Unit/Lens/ModelSync", test_lens_model_sync);

    Test.run ();

    return 0;
  }

  static Unity.Lens exported_lens;
  static bool name_owned = false;

  public static void test_lens_export ()
  {
    // register us a name on the bus
    Bus.own_name (BusType.SESSION, "com.canonical.Unity.Lens.Test", 0,
                  () => {}, () => { name_owned = true; },
                  () => { debug ("Name lost"); assert_not_reached (); });

    var lens = new Unity.Lens ("/com/canonical/Unity/Lens/Test",
                               "unity_lens_test");
    lens.search_in_global = false;
    lens.search_hint = "Search hint";
    lens.export ();

    exported_lens = lens;
  }

  static Unity.Scope local_scope;

  public static void test_local_scope_init ()
  {
    assert (exported_lens != null);

    var scope = new Unity.Scope ("/com/canonical/Unity/LocalScope/Test");

    local_scope = scope;
    exported_lens.add_local_scope (scope);
  }

  public static void test_remote_scope_init ()
  {
    assert (exported_lens != null);

    bool scope_up = false;

    // the remote scope doesn't have a dbus service file installed, so we
    // expect that something (dbus-test-runner) started it already
    var ml = new MainLoop ();
    uint watch_id = Bus.watch_name (BusType.SESSION, 
                                    "com.canonical.Unity.Scope0.Test", 0,
                                    () => { scope_up = true; ml.quit (); },
                                    () => { scope_up = false; });

    Timeout.add (2000, () => { ml.quit (); return false; });
    ml.run ();

    assert (scope_up == true);

    flush_bus ();
    // we need to wait a bit more to connect to the proxy
    // FIXME: find a better way to do this
    ml = new MainLoop ();
    Timeout.add (500, () => { ml.quit (); return false; });
    ml.run ();

    Bus.unwatch_name (watch_id);
    // should be still up
    assert (scope_up == true);
  }

  private static void call_lens_search (string search_string, Func<Variant?>? cb = null)
  {
    var bus = Bus.get_sync (BusType.SESSION);

    var vb = new VariantBuilder (new VariantType ("(sa{sv})"));
    vb.add ("s", search_string);
    vb.open (new VariantType ("a{sv}"));
    vb.close ();

    bus.call.begin ("com.canonical.Unity.Lens.Test",
                    "/com/canonical/Unity/Lens/Test",
                    "com.canonical.Unity.Lens",
                    "Search",
                    vb.end (),
                    null,
                    0,
                    -1,
                    null,
                    (obj, res) =>
    {
      try
      {
        var reply = bus.call.end (res);
        if (cb != null) cb (reply);
      }
      catch (Error e)
      {
        warning ("%s", e.message);
      }
    });
  }

  public static void test_lens_search ()
  {
    var ml = new MainLoop ();
    // make sure we got response from own_name, so we can send ourselves
    // a dbus method call
    Idle.add (() =>
    {
      if (name_owned) ml.quit ();
      return !name_owned;
    });

    ml.run ();

    call_lens_search ("foo");
  }

  public static void test_local_scope_search ()
  {
    assert (local_scope != null);

    var ml = new MainLoop ();
    bool got_search_changed = false;
    ulong sig_id = local_scope.search_changed.connect ((lens_search) =>
    {
      assert (lens_search.search_string == "foo");
      got_search_changed = true;
      lens_search.finished ();
      ml.quit ();
    });

    // wait for the signal or timeout
    Timeout.add (1000, () => { ml.quit (); return false; });
    ml.run ();

    assert (got_search_changed == true);
    SignalHandler.disconnect (local_scope, sig_id);
  }

  private static void flush_bus ()
  {
    var bus = Bus.get_sync (BusType.SESSION);
    bus.flush_sync ();

    var ml = new MainLoop ();
    Idle.add (() => { ml.quit (); return false; });
    ml.run ();
    // this should flush the dbus method calls
  }

  public static void test_lens_return_after_scope_finish ()
  {
    assert (local_scope != null);

    var ml = new MainLoop ();
    bool got_search_changed = false;
    bool finish_called = false;

    Func<Variant?> cb = () =>
    {
      ml.quit ();
    };
    ulong sig_id = local_scope.search_changed.connect ((lens_search) =>
    {
      got_search_changed = true;
      Timeout.add (750, () =>
      {
        finish_called = true;
        lens_search.finished ();
        return false;
      });
    });

    // we want to make sure the Search DBus call doesn't return before we
    // call finished on the LensSearch instance
    call_lens_search ("qoo", cb);
    Timeout.add (5000, () => { ml.quit (); return false; });
    ml.run ();

    assert (got_search_changed == true);
    assert (finish_called == true);

    SignalHandler.disconnect (local_scope, sig_id);
  }

  public static void test_lens_two_searches ()
  {
    assert (local_scope != null);

    var ml = new MainLoop ();
    Cancellable? canc1 = null;
    Cancellable? canc2 = null;
    bool got_search_changed = false;
    uint finish_calls = 0;

    ulong sig_id = local_scope.search_changed.connect ((lens_search, search_type, cancellable) =>
    {
      got_search_changed = true;
      switch (lens_search.search_string)
      {
        case "foo1": canc1 = cancellable; break;
        case "foo2": canc2 = cancellable; break;
        default: assert_not_reached ();
      }

      var timer_id = Timeout.add (1000, () =>
      {
        finish_calls++;
        lens_search.finished ();

        if (finish_calls == 2) ml.quit ();
        return false;
      });

      ml.quit ();
    });

    string order = "";
    var reply_ml = new MainLoop ();
    int replies = 0;
    Func<Variant?> foo1_finished_cb = () =>
    {
      order += "1";
      if (++replies == 2) reply_ml.quit ();
    };
    Func<Variant?> foo2_finished_cb = () =>
    {
      order += "2";
      if (++replies == 2) reply_ml.quit ();
    };

    // we dont want to wait indefinitely
    var bad_timer = Timeout.add (2000, () => { assert_not_reached (); });
    call_lens_search ("foo1", foo1_finished_cb);
    ml.run ();
    flush_bus ();
    ml = new MainLoop ();
    call_lens_search ("foo2", foo2_finished_cb);
    ml.run ();

    Source.remove (bad_timer);

    assert (canc1 != null);
    assert (canc2 != null);

    assert (canc1.is_cancelled () == true);
    assert (canc2.is_cancelled () == false);

    flush_bus ();

    // the timers are still running and we need to wait for the replies
    reply_ml.run ();

    // make sure the first search finished earlier that the second
    assert (order == "12");

    SignalHandler.disconnect (local_scope, sig_id);
  }

  public static void test_lens_model_sync ()
  {
    assert (local_scope != null);

    bool got_search_changed = false;
    var ml = new MainLoop ();
    ulong sig_id = local_scope.search_changed.connect ((lens_search, search_type, cancellable) =>
    {
      assert (lens_search.search_string == "model-sync");
      got_search_changed = true;

      Timeout.add (300, () =>
      {
        var model = lens_search.results_model;
        model.append ("uri", "icon", 0, "mimetype", "display-name",
                      "comment", "dnd-uri");
        lens_search.finished ();
        return false;
      });
    });

    uint64 seqnum = 0;
    call_lens_search ("model-sync", (reply) =>
    {
      assert (reply != null);
      HashTable<string, Variant> ht = (HashTable<string, Variant>) reply.get_child_value (0);
      unowned Variant? seqnum_v = ht.lookup ("model-seqnum");
      assert (seqnum_v != null);
      seqnum = seqnum_v.get_uint64 ();
      ml.quit ();
    });
    Timeout.add (3000, () => { assert_not_reached (); });
    ml.run ();
    // libunity will emit warnings if the models are out-of-sync, those would
    // fail this test

    assert (got_search_changed == true);
    // FIXME: not too great test if previous tests did something with the model
    assert (seqnum > 0);

    SignalHandler.disconnect (local_scope, sig_id);
  }
}
